package kasapp

import (
	"bytes"
	"errors"
	"log/slog"
	"runtime/pprof"
	"time"

	"github.com/getsentry/sentry-go"
)

type stacktraceCollector struct {
	log      *slog.Logger
	hub      *sentry.Hub
	interval time.Duration
	start    time.Time
}

func startStacktraceCollector(log *slog.Logger, hub *sentry.Hub, interval time.Duration) {
	s := &stacktraceCollector{log: log, hub: hub, interval: interval, start: time.Now().UTC()}
	s.Start()
}

func (s *stacktraceCollector) Start() {
	time.AfterFunc(s.interval, s.capture)
}

func (s *stacktraceCollector) capture() {
	defer s.Start() // restart for periodic capturing

	delayedBySec := time.Since(s.start) / time.Second

	s.log.Warn("Capturing goroutines because of delayed shutdown", slog.Time("start", s.start), slog.Int64("delayed_by_sec", int64(delayedBySec)))

	var b bytes.Buffer
	_ = pprof.Lookup("goroutine").WriteTo(&b, 1)

	e := sentry.NewEvent()
	e.Level = sentry.LevelWarning
	e.SetException(errors.New("delayed shutdown"), -1)
	e.Attachments = []*sentry.Attachment{
		{
			Filename:    "goroutines.txt",
			ContentType: "text/plain",
			Payload:     b.Bytes(),
		},
	}
	e.Extra = map[string]any{
		"start":          s.start,
		"delayed_by_sec": delayedBySec,
	}
	s.hub.CaptureEvent(e)
}
