package ioz

import (
	"crypto/ed25519"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"io"
	"log/slog"
	"os"

	"google.golang.org/grpc/mem"
)

const (
	// hmacSHA256MinKeySize is the minimum HMAC SHA256 key size that is considered secure.
	// See https://www.rfc-editor.org/rfc/rfc2104#section-3 and https://en.wikipedia.org/wiki/HMAC.
	hmacSHA256MinKeySize = sha256.Size
	// hmacSHA256MaxKeySize is the maximum HMAC SHA256 key size that is not hashed into 32 bytes.
	hmacSHA256MaxKeySize = sha256.BlockSize

	readAllBufSize = 32 * 1024
)

// DiscardData discards up to 8KiB of data from r.
// Function can be used to drain HTTP client response body.
// See https://pkg.go.dev/net/http#Response.
func DiscardData(r io.Reader) error {
	_, err := io.Copy(io.Discard, io.LimitReader(r, 8*1024))
	if err != nil {
		return fmt.Errorf("failed to read data: %w", err)
	}
	return nil
}

func LoadSHA256Base64Secret(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, hmacSHA256MinKeySize, hmacSHA256MaxKeySize)
}

func LoadEd25519Base64PublicKey(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, ed25519.PublicKeySize, ed25519.PublicKeySize)
}

func LoadBase64Secret(log *slog.Logger, filename string, minLen, maxLen int) ([]byte, error) {
	encodedAuthSecret, err := os.ReadFile(filename) //nolint: gosec
	if err != nil {
		return nil, fmt.Errorf("read file: %w", err)
	}
	decodedAuthSecret := make([]byte, base64.StdEncoding.DecodedLen(len(encodedAuthSecret)))

	n, err := base64.StdEncoding.Decode(decodedAuthSecret, encodedAuthSecret)
	if err != nil {
		return nil, fmt.Errorf("decoding: %w", err)
	}

	if n < minLen || n > maxLen {
		// TODO make this an error in 18.0, remove logging
		// TODO https://gitlab.com/gitlab-org/cluster-integration/gitlab-agent/-/issues/617
		log.Warn(fmt.Sprintf("secret in file %s is %d bytes long; allowed length is from %d to %d bytes", filename, n, minLen, maxLen))
		//return nil, fmt.Errorf("secret is %d bytes long; allowed length is from %d to %d bytes", n, minLen, maxLen)
	}

	return decodedAuthSecret[:n], nil
}

// ReadAll reads from the reader until io.EOF or an error.
// Returns the read data as a slice of buffers or an error.
func ReadAll(r io.Reader) (mem.BufferSlice, error) {
	return ReadAllWithPool(r, mem.DefaultBufferPool())
}

func ReadAllWithPool(r io.Reader, pool mem.BufferPool) (mem.BufferSlice, error) {
	var result mem.BufferSlice
	wt, ok := r.(io.WriterTo)
	if ok {
		// This is more optimal since wt knows the size of chunks it wants to write and, hence, we can allocate
		// buffers of an optimal size to fit them. E.g. might be a single big chunk, and we wouldn't chop it into pieces.
		w := mem.NewWriter(&result, pool)
		_, err := wt.WriteTo(w)
		if err != nil {
			result.Free() // free any written buffers
			return nil, err
		}
		return result, nil
	}
	for {
		buf := pool.Get(readAllBufSize)
		// We asked for 32KiB but may have been given a bigger buffer. Use all of it if that's the case.
		*buf = (*buf)[:cap(*buf)]
		usedCap := 0
		for {
			n, err := r.Read((*buf)[usedCap:])
			usedCap += n
			if err != nil {
				if err == io.EOF {
					if usedCap == 0 {
						// Nothing in this buf, put it back
						pool.Put(buf)
					} else {
						*buf = (*buf)[:usedCap]
						buffer := mem.NewBuffer(buf, pool)
						result = append(result, buffer)
					}
					return result, nil
				}
				pool.Put(buf)
				result.Free()
				return nil, err
			}
			if len(*buf) == usedCap {
				buffer := mem.NewBuffer(buf, pool)
				result = append(result, buffer)
				break // grab a new buf from pool
			}
		}
	}
}

// ReadAllFunc fully reads the reader and passes the data to the callback.
// Callback must not retain the buffer or any portion of it.
func ReadAllFunc(r io.Reader, f func([]byte) error) error {
	bs, err := ReadAll(r)
	if err != nil {
		return fmt.Errorf("read all: %w", err)
	}
	data := bs.MaterializeToBuffer(mem.DefaultBufferPool())
	bs.Free()
	defer data.Free()
	return f(data.ReadOnlyData())
}
